import logging

from .core import AuthenticationException

logger = logging.getLogger(__name__)


class UsernamePasswordAuthenticationToken(object):

    def __init__(self, username, password):
        if not username:
            raise AuthenticationException("Username cannot be None.")
        self.username = username
        self.password = password


class UserTokenAuthenticationToken(object):

    def __init__(self, token):
        self.token = token

    @property
    def user_uid(self):
        return self.token.split(":")[0]

    @property
    def security_token(self):
        return self.token.split(":")[1]


class AbstractUserTokenAuthenticationProvider(object):
    """This provides a very basic shared key authentication system.
    """

    def supports(self, authentication_token):
        if authentication_token and isinstance(authentication_token, UserTokenAuthenticationToken):
            return True
        else:
            return False

    def authenticate(self, authentication_token):

        if not authentication_token.user_uid or not authentication_token.security_token:
            raise AuthenticationException("Invalid authentication token.")

        self._authenticate_user(authentication_token)

    def _authenticate_user(self, authentication_token):
        raise NotImplementedError





import functools
import inspect
import json
import logging
import traceback
import urllib
import urllib.parse

from collections import namedtuple

from decorator import decorator

from tornado.escape import to_basestring
from tornado.web import RequestHandler, HTTPError

from dorthy import template
from dorthy.dp import ObjectDict
from dorthy.enum import DeclarativeEnum
from dorthy.json import jsonify
from dorthy.session import session_store, Session
from dorthy.request import WebRequestHandlerProxyMixin
from dorthy.security import SecurityManager, AccessDeniedError, AuthenticationException
from dorthy.settings import config
from dorthy.utils import camel_decode, native_str


logger = logging.getLogger(__name__)

ErrorResponse = namedtuple("ErrorResponse",
                           ["status_code", "message", "exception", "stack"])


class MediaTypes(DeclarativeEnum):

    HTML = "text/html"
    JSON = "application/json"


def consumes(media=MediaTypes.JSON, arg_name="model", underscore_case=True):

    def _process_camel_case(d):
        processed = dict()
        for key, value in d.items():
            processed[camel_decode(key)] = value
        return processed

    def _parse_json(request):
        s = to_basestring(request.body)
        if underscore_case:
            json_dict = json.loads(s, object_hook=_process_camel_case)
        else:
            json_dict = json.loads(s)
        return ObjectDict(json_dict)

    def _consumes(f, handler, *args, **kwargs):

        # check for proper content type
        if not handler.request.headers.get("Content-Type", "").startswith(media.value):
            raise HTTPError(400, "Invalid Content-Type received.")

        if media == MediaTypes.JSON:
            # check keyword args first
            if arg_name in kwargs:
                kwargs[arg_name] = _parse_json(handler.request)
            else:
                sig = inspect.signature(f)
                params = sig.parameters
                for indx, (name, param) in enumerate(params.items()):
                    if name == arg_name or \
                            (param.annotation != inspect.Parameter.empty and param.annotation == "model"):
                        args = list(args)
                        args[indx - 1] = _parse_json(handler.request)
                        break

                    # model param not contained in method signature
                    if indx == len(args):
                        raise TypeError("No model argument found in method signature")
        else:
            raise HTTPError(500, "MediaType not supported.")

        return f(handler, *args, **kwargs)

    return decorator(_consumes)


def produces(media=MediaTypes.JSON, root=None, camel_case=True, ignore_attributes=None):

    def _produces(f, handler, *args, **kwargs):
        handler.media_type = media
        handler.set_header("Content-Type", media.value)
        val = f(handler, *args, **kwargs)
        if val and not handler.finished:
            if media == MediaTypes.JSON:
                if root is None and "produces_wrapper" in handler.application.settings:
                    root_wrapper = handler.application.settings["produces_wrapper"]
                else:
                    root_wrapper = root
                handler.write(jsonify(val,
                                      root=root_wrapper,
                                      camel_case=camel_case,
                                      ignore_attributes=ignore_attributes))
            elif media == MediaTypes.HTML:
                handler.write(val)
        return None

    return decorator(_produces)


@decorator
def render(f, handler, *args, **kwargs):
    val = f(handler, *args, **kwargs)
    if val and not handler.finished:
        if isinstance(val, str):
            handler.render(val)
        else:
            handler.render(val[0], **val[1])


@decorator
def redirect(f, handler, *args, **kwargs):
    val = f(handler, *args, **kwargs)
    if val and not handler.finished:
        handler.redirect(val)


class BaseHandler(RequestHandler, WebRequestHandlerProxyMixin):

    SESSION_COOKIE_KEY = "s"

    DEFAULT_SESSION_TIMEOUT = 1800
    if "web.session_timeout" in config:
        DEFAULT_SESSION_TIMEOUT = config.web.get("session_timeout", 1800)

    COOKIE_DOMAIN = None
    if "web.cookie_domain" in config:
        COOKIE_DOMAIN = config.web.cookie_domain

    USE_SECURE_COOKIE = True if "web.cookie_secret" in config and config.web.enabled("cookie_secret") else False

    def __init__(self, application, request, **kwargs):
        self.media_type = MediaTypes.HTML
        self.application = application
        self._request_finished = False
        self.__session = None
        self.__debug = "debug" in self.application.settings and \
                       self.application.settings["debug"]
        self.__client_ip = None

        # initialize framework template system -- replace tornado's
        if "template_conf" in self.application.settings:
            self.require_setting("template_path", feature="@dorthy.web.template.engine")
            conf = dict(self.application.settings["template_conf"])
            conf["auto_reload"] = self.__debug
            template.config_environment(self.get_template_path(), **conf)
            self.__use_framework_templates = True

        super().__init__(application, request, **kwargs)

    @property
    def client_ip(self):
        """
        Provides a method to retrieve the client ip address
        behind a load balancer (ELB)
        """
        if not self.__client_ip:
            ip = self.request.headers.get("X-Forwarded-For", self.request.remote_ip)
            self.__client_ip = ip.split(",")[0].strip()
        return self.__client_ip

    @property
    def debug(self):
        return self.__debug

    @property
    def finished(self):
        return self._finished

    def get_user_agent(self):
        return self.request.headers.get("User-Agent")

    def get_current_user(self):
        if SecurityManager().authenticated():
            return SecurityManager().get_principal()
        else:
            return None

    def get_user_locale(self):
        principal = self.get_current_user()
        if principal is not None:
            return principal.locale
        else:
            return None

    def set_secure_cookie(self, name, value, expires_days=None,
                          domain=None, expires=None, path="/", **kwargs):

        if domain is None and self.COOKIE_DOMAIN is not None:
            domain = self.COOKIE_DOMAIN

        super().set_secure_cookie(name,
                                  value,
                                  expires_days=expires_days,
                                  domain=domain,
                                  expires=expires,
                                  path=path,
                                  **kwargs)

    def set_nocache_headers(self):
        self.set_header('Cache-Control', 'no-cache, no-store, must-revalidate')
        self.set_header('Pragma', 'no-cache')
        self.set_header('Expires', 0)

    def clear_session(self):
        session = self.get_session()
        if session is not None:
            session.invalidate()

    def get_session(self, create=False, timeout=DEFAULT_SESSION_TIMEOUT, update_access=True):
        if self.__session is None:
            session_id = self.__get_session_cookie()
            if session_id:
                self.__session = session_store.load(native_str(session_id))
            if self.__session is None and create:
                self.__session = Session(session_store.generate_session_id(),
                                         timeout=timeout,
                                         update_access=update_access)
                self.__set_session_cookie()
        return self.__session

    def __get_session_cookie(self):
        if self.USE_SECURE_COOKIE:
            return self.get_secure_cookie(self.SESSION_COOKIE_KEY)
        else:
            return self.get_cookie(self.SESSION_COOKIE_KEY)

    def __set_session_cookie(self):
        if self.__session is not None:
            if self.USE_SECURE_COOKIE:
                self.set_secure_cookie(self.SESSION_COOKIE_KEY, self.__session.session_id)
            else:
                self.set_cookie(self.SESSION_COOKIE_KEY, self.__session.session_id)
        else:
            logger.warn("Set Session cookie called for empty session.")

    def __save_session(self):
        # load the session if it exists so that the session
        # store can update its timestamp / expiration period
        session = self.get_session()
        if session is not None:
            session_store.save(session)
            if not session.valid:
                self.clear_cookie(self.SESSION_COOKIE_KEY)
        elif self.__get_session_cookie():
            self.clear_cookie(self.SESSION_COOKIE_KEY)

    def on_finish(self):
        pass

    def render(self, template_name, **kwargs):
        """
        Renders the given template using the passed keyword args

        :param template_name: the template name
        :param kwargs: keyword args passed to the template
        """
        if self.__use_framework_templates:
            temp = template.get_template(self.get_template_path(), template_name)
            namespace = dict(
                handler=self,
                request=self.request,
                current_user=self.current_user,
                locale=self.locale,
                static_url=self.static_url,
                xsrf_form_html=self.xsrf_form_html,
                reverse_url=self.reverse_url
            )
            namespace.update(kwargs)
            rendered = temp.render(namespace)
            self.finish(rendered)
        else:
            super().render(template_name, kwargs)

    def finish(self, chunk=None):
        if not self._request_finished:
            # prevents a recursive loop on finish if exception raised
            self._request_finished = True
            self.__save_session()
            self.on_finish()
        super().finish(chunk)

    def write_error(self, status_code, **kwargs):

        e = stack = None
        exc_info = kwargs.get("exc_info", None)

        if status_code == 401:
            message = "User not authorized."
        elif status_code == 403:
            message = "User forbidden."
        else:
            if exc_info:
                if isinstance(exc_info[1], AccessDeniedError):
                    status_code = 403
                    message = "User forbidden."
                    self.set_status(403, message)
                elif isinstance(exc_info[1], AuthenticationException):
                    status_code = 401
                    message = "User not authorized."
                    self.set_status(401, message)
                else:
                    t, e, tb = exc_info
                    stack = traceback.format_tb(tb)
                    message = str(e) if self.debug else "An internal server error occurred."
            else:
                message = "An unknown error occurred."

        error = ErrorResponse(status_code=status_code,
                              message=message,
                              exception=e.__class__.__name__ if self.debug and e else None,
                              stack=stack if self.debug and stack else None)

        if self.media_type == MediaTypes.JSON:
            self.set_header("Content-Type", MediaTypes.JSON.value)
            self.write(jsonify(error._asdict(), "error"))
        else:
            if self.debug:
                self.render("error/error-dev.html", error=error)
            else:
                self.render("error/error.html", error=error)


class TemplateHandler(BaseHandler):

    def initialize(self, template, status=None):
        self.template = template
        self.status = status

    def get(self):
        if self.status:
            self.set_status(self.status)
        self.render(self.template)


def authenticated(redirect=False):

    def _authenticated(f, handler, *args, **kwargs):
        if not SecurityManager().authenticated():
            SecurityManager().load_context(handler)
            if not SecurityManager().authenticated():
                if redirect and handler.request.method in ("GET", "POST", "HEAD"):
                    url = handler.get_login_url()
                    if "?" not in url:
                        if urllib.parse.urlsplit(url).scheme:
                            next_url = handler.request.full_url()
                        else:
                            next_url = handler.request.uri
                        url += "?" + urllib.parse.urlencode(dict(next=next_url))
                    handler.redirect(url)
                    return
                raise HTTPError(401, "User not authorized.")

        return f(handler, *args, **kwargs)

    return decorator(_authenticated)



import inspect
import logging
import weakref

from collections import MutableMapping


logger = logging.getLogger(__name__)


class Singleton(object):
    """
    A decorated to implement the singleton pattern.
    This implementation is not thread safe.

    Based on: http://stackoverflow.com/questions/42558/python-and-the-singleton-pattern
    """

    def __init__(self, decorated_class):
        self.__decorated_class = decorated_class
        self.__doc__ = decorated_class.__doc__

    def __call__(self):
        """
        Returns the singleton instance.
        """
        try:
            return self.__instance
        except AttributeError:
            self.__instance = self.__decorated_class()
            return self.__instance


class MultipleObservableErrors(Exception):

    def __init__(self, message, errors=None):
        super().__init__(message)
        self.__errors = errors

    @property
    def errors(self):
        return self.__errors


class Observable(object):
    """Provides an observable class the uses weak references so
    that objects are not held in memory by the observable.
    """

    def __init__(self, handle_errors=True):
        self.__listeners = dict()
        self.__handle_errors = handle_errors

    def __call__(self, *args, **kwargs):
        """Calls the listeners with the given arguments and
        keyword arguments.
        """
        if not self.__listeners:
            return

        errors = list()
        listeners = dict()
        for k, v in self.__listeners.items():
            call_obj = None
            if v[0] == "f":
                call_obj = v[2]
            elif v[0] == "m":
                obj = v[2]()
                if obj:
                    call_obj = getattr(obj, v[3].__name__)
            else:
                raise ValueError("Only supports functions or methods as listeners.")

            # execute callable
            if call_obj:
                listeners[v[1]] = v
                try:
                    call_obj(*args, **kwargs)
                except Exception as e:
                    if self.__handle_errors:
                        logger.exception("Failed to execute observable callback.")
                    else:
                        # collect the errors
                        errors.append(e)

        # clean out references that are no longer valid
        self.__listeners = listeners

        # raise exceptions generated
        if errors:
            raise MultipleObservableErrors("Failed to execute observable callbacks.", errors)

    def __contains__(self, item):
        if inspect.isfunction(item) or inspect.ismethod(item):
            return id(item) in self.__listeners
        else:
            return item in self.__listeners

    def register(self, listener):
        """Registers a listener with the observable class.  The
        listener must either be a method or a function.

        returns: the listener id that can be used to remove the listener
        """
        ref_key = id(listener)
        if inspect.isfunction(listener):
            ref = ("f", ref_key, listener)
        elif inspect.ismethod(listener):
            weak_obj = weakref.ref(listener.__self__)
            ref = ("m", ref_key, weak_obj, listener.__func__)
        else:
            raise ValueError("Only supports functions or methods as listeners.")

        self.__listeners[ref_key] = ref

        return ref_key

    def remove(self, listener_id):
        """Removes the listener for the given id.

        returns: True if the listener has been found and removed, otherwise False
        """
        return True if self.__listeners.pop(listener_id, None) else False


class ObjectMap(MutableMapping):
    """Converts object attribute access to a
    map view.
    """

    def __init__(self, obj, ignore_=True):
        self.__object = obj
        self.__ignore_ = ignore_

    def _get_attrs(self):
        if self.__ignore_:
            return [d for d in dir(self.__object) if not d.startswith("_")]
        else:
            return dir(self.__object)

    def _check_key(self, key):
        if self.__ignore_ and key.startswith("_"):
            raise KeyError("Invalid attribute: " + key)

    def __contains__(self, key):
        self._check_key(key)
        return hasattr(self.__object, key)

    def __len__(self):
        return len(self._get_attrs())

    def __getitem__(self, key):
        self._check_key(key)
        try:
            return getattr(self.__object, key)
        except AttributeError:
            raise KeyError(key)

    def __delitem__(self, key):
        self._check_key(key)
        try:
            delattr(self.__object, key)
        except AttributeError:
            raise KeyError(key)

    def __setitem__(self, key, value):
        self._check_key(key)
        setattr(self.__object, key, value)

    def __iter__(self):
        return iter(self._get_attrs())


class ObjectDict(dict):
    """Extends dict to provide object attribute get and set access.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        self[name] = value



from .base import BaseSessionStore, Session, DEFAULT_SESSION_TIMEOUT

from dorthy import redis
from dorthy.settings import config


WEB_SESSION_PREFIX = "web:session"


class RedisSessionStore(BaseSessionStore):

    @staticmethod
    def _store_key(session_id):
        return redis.create_key(WEB_SESSION_PREFIX, session_id)

    def load(self, session_id):
        session_data = redis.client.get(self._store_key(session_id))
        return self._validate_session(Session.decode(session_data)) if session_data else None

    def _delete(self, session_id):
        redis.client.delete(self._store_key(session_id))

    def _store_session(self, session):
        if session.valid:
            session._update_accessed()
            session_data = session.encode()
            timeout = config.web.session_timeout if session.timeout <= 0 else session.timeout
            if timeout <= 0:
                # always use a timeout with redis store to prevent orphan data
                timeout = DEFAULT_SESSION_TIMEOUT

            # timeout is not correct for non-updating sessions
            # will be caught by load method and expired check
            redis.client.setex(self._store_key(session.session_id),
                               timeout,
                               session_data)
        else:
            self._delete(session.session_id)


import base64
import bcrypt
import codecs
import hashlib
import hmac
import string
import sys

from functools import partial
from hashlib import sha1

from Crypto.Cipher import AES
from Crypto import Random
from Crypto.Random import random

from dorthy.enum import DeclarativeEnum
from dorthy.settings import config


def generate_id(chars=string.ascii_uppercase + string.digits, size=8):
    return ''.join(random.choice(chars) for x in range(size))


def sign_message(key, msg, digest=sha1):
    hashed = hmac.new(key, msg, digest)
    return base64.b64encode(hashed.digest())


def aes_encrypt(key, text, base64_encode=True, encoding=None):
    """
    Provides an AES encryption routine using cipher feedback
    block algorithm

    :rtype : encrypted bytes
    """
    assert key is not None

    iv = Random.new().read(AES.block_size)
    cipher = AES.new(key, AES.MODE_CFB, iv)
    # encode string into bytes
    if not encoding:
        encoding = sys.getdefaultencoding()
    value = iv + cipher.encrypt(text.encode(encoding))
    if base64_encode:
        value = base64.b64encode(value).decode(encoding)
    return value


def aes_decrypt(key, cipher_text, base64_encode=True, encoding=None):
    """
    Provides an AES decryption routine using cipher feedback
    block algorithm

    :type key: the decrypted text
    """
    assert key is not None

    iv = cipher_text[0:AES.block_size]
    cipher = AES.new(key, AES.MODE_CFB, iv)
    if not encoding:
        encoding = sys.getdefaultencoding()
    if isinstance(cipher_text, str):
        # encode into bytes
        cipher_text = cipher_text.encode(encoding)
    if base64_encode:
        cipher_text = base64.b64decode(cipher_text)
    text = cipher.decrypt(cipher_text)
    return text[AES.block_size:].decode(encoding)


if "security.encryption_key" in config:
    private_key = config.security.encryption_key.encode(sys.getdefaultencoding())
else:
    private_key = None

encrypt = partial(aes_encrypt, private_key)
decrypt = partial(aes_decrypt, private_key)


class PasswordEncryptionAlgorithms(DeclarativeEnum):

    BCrypt = "1",
    SHA1 = "2",


def encrypt_password(password, encryption_algorithm=PasswordEncryptionAlgorithms.BCrypt):

    assert password, "Password cannot be None"

    if encryption_algorithm == PasswordEncryptionAlgorithms.BCrypt:
        return bcrypt.hashpw(password, bcrypt.gensalt()) if password else None
    elif encryption_algorithm == PasswordEncryptionAlgorithms.SHA1:
        m = hashlib.sha1()
        m.update(password.encode('utf-8'))
        return codecs.encode(m.digest(), 'hex_codec').decode("utf-8")
    else:
        raise NotImplemented


def validate_password(password, token, encryption_algorithm=PasswordEncryptionAlgorithms.BCrypt):
    if encryption_algorithm == PasswordEncryptionAlgorithms.BCrypt:
        return bcrypt.hashpw(password, token) == token
    else:
        raise NotImplemented



from os import path

from jinja2 import Environment, FileSystemLoader

__template_environments = dict()


def get_template(template_path, template_name, parent=None):
    template_env = get_environment(template_path)
    return template_env.get_template(template_name, parent=parent)


def get_environment(template_path):
    if template_path not in __template_environments:
        raise LookupError("Template environment not configured.")
    return __template_environments[template_path]


def config_environment(template_path,
                       block_start_string="{%",
                       block_end_string="%}",
                       variable_start_string="{{",
                       variable_end_string="}}",
                       comment_start_string="{#",
                       comment_end_string="#}",
                       extensions=None,
                       auto_escape=True,
                       cache_size=-1,
                       auto_reload=True,
                       global_vars=None):

    def _auto_escape(template_name):
        if template_name is None or "." not in template_name:
            return False
        ext = template_name.rsplit(".", 1)[1]
        return ext in ("html", "htm", "xml")

    if template_path in __template_environments:
        return

    abspath = path.abspath(template_path)
    loader = FileSystemLoader(abspath)

    if auto_escape:
        if extensions is not None:
            exts = list(extensions)
            exts.append("jinja2.ext.autoescape")
        else:
            exts = ["jinja2.ext.autoescape"]

        env = Environment(autoescape=_auto_escape,
                          block_start_string=block_start_string,
                          block_end_string=block_end_string,
                          variable_start_string=variable_start_string,
                          variable_end_string=variable_end_string,
                          comment_start_string=comment_start_string,
                          comment_end_string=comment_end_string,
                          extensions=exts,
                          cache_size=cache_size,
                          auto_reload=auto_reload,
                          loader=loader)
    else:
        env = Environment(block_start_string=block_start_string,
                          block_end_string=block_end_string,
                          variable_start_string=variable_start_string,
                          variable_end_string=variable_end_string,
                          comment_start_string=comment_start_string,
                          comment_end_string=comment_end_string,
                          extensions=extensions,
                          cache_size=cache_size,
                          auto_reload=auto_reload,
                          loader=loader)

    if global_vars:
        env.globals.update(global_vars)

    __template_environments[template_path] = env


def render(template_path, template_name, parent=None, **kwargs):
    template = get_template(template_path, template_name, parent=parent)
    return template.render(**kwargs)



import logging
import urllib.parse

from .auth import UsernamePasswordAuthenticationToken
from .core import AuthenticationException, SecurityManager

from dorthy.web import BaseHandler, render, redirect

logger = logging.getLogger(__name__)


class UserPasswordAuthHandler(BaseHandler):

    def _get_default_url(self):
        if "default_url" in self.application.settings:
            return self.application.settings["default_url"]
        else:
            return "/"

    @render
    def get(self):
        return "login.html", dict(next=self.get_argument("next", self._get_default_url()))

    @redirect
    def post(self):

        username = self.get_argument("username")
        password = self.get_argument("password")
        next_url = self.get_argument("next", self._get_default_url())

        try:
            token = UsernamePasswordAuthenticationToken(username, password)
            auth_provider = SecurityManager().get_authentication_provider(token)
            if not auth_provider:
                raise AuthenticationException("No authentication provider found for user authentication.")

            auth_provider.authenticate(token)
            if not SecurityManager().authenticated():
                raise AuthenticationException()

            SecurityManager().store_context(self)
        except AuthenticationException:
            logger.exception("Failed to authenticate username: %s", username)
            return self.request.path + "?" + urllib.parse.urlencode(dict(next=next_url))
        else:
            return next_url


class DefaultLogoutHandler(BaseHandler):

    @redirect
    def get(self):
        self.clear_session()
        return "/auth/login"
